import random
import sys
import os
import glob
from datetime import datetime

import pandas as pd
import numpy as np
import torch
import cv2
import time
import urllib.error
import serial
import adafruit_fingerprint
from PIL import Image


def Ticks():
    return int(time.time() * 1000)


class ImgRex:  # YoloV3 dan V4
    def __init__(self):
        self.output_layers = None
        self.net = None
        self.colors = None
        self.classes = None
        self.classes_path = None

    def __map(self, x, inMin, inMax, outMin, outMax):
        return (x - inMin) * (outMax - outMin) // (inMax - inMin) + outMin

    def load(self, weight_path, cfg, classes):
        self.classes = None
        self.classes_path = classes
        with open(classes, "r") as f:
            self.classes = [line.strip() for line in f.readlines()]
        self.colors = np.random.uniform(
            0, 255, size=(len(self.classes), 3))  # optional
        self.net = cv2.dnn.readNet(weight_path, cfg)
        # self.net = cv2.dnn.readNetFromDarknet(cfg, weight_path)
        layer_names = self.net.getLayerNames()
        self.output_layers = [layer_names[i - 1]
                              for i in self.net.getUnconnectedOutLayers()]
        # self.output_layers = self.net.getUnconnectedOutLayersNames()

    @staticmethod
    def draw(frame, detection):
        if detection is not []:
            for idx in detection:
                color = idx["color"]
                cv2.rectangle(
                    frame, (idx["x"], idx["y"]), (idx["x"] + idx["width"], idx["y"] + idx["height"]), color, 2)
                tl = round(0.002 * (frame.shape[0] + frame.shape[1]) / 2) + 1
                c1, c2 = (int(idx["x"]), int(idx["y"])), (int(
                    idx["width"]), int(idx["height"]))

                tf = int(max(tl - 1, 1))  # font thickness
                t_size = cv2.getTextSize(
                    idx["class"], 0, fontScale=tl / 3, thickness=tf)[0]
                c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3

                cv2.rectangle(frame, c1, c2, color, -1, cv2.LINE_AA)  # filled
                cv2.putText(frame, idx["class"] + " " + str(int(idx["confidence"] * 100)) + "%", (c1[0], c1[1] - 2), 0,
                            tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)
                # cv2.putText(frame, idx["class"] + " " + str(int(random.randint(65, 75))) + "%", (c1[0], c1[1] - 2), 0,tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)
                cv2.circle(frame, (
                    int(idx["x"] + int(idx["width"] / 2)), int(idx["y"] + int(idx["height"] / 2))),
                           4, color, -1)
                cv2.putText(frame, str(int(idx["x"] + int(idx["width"] / 2))) + ", " + str(
                    int(idx["y"] + int(idx["height"] / 2))), (
                                int(idx["x"] + int(idx["width"] / 2) + 10),
                                int(idx["y"] + int(idx["height"] / 2) + 10)), cv2.FONT_HERSHEY_PLAIN, tl / 2,
                            [255, 255, 255], thickness=tf, lineType=cv2.LINE_AA)
        return frame

    def predict(self, frame):
        blob = cv2.dnn.blobFromImage(
            frame, 0.00392, (416, 416), (0, 0, 0), True, crop=False)
        height, width, ch = frame.shape
        self.net.setInput(blob)
        outs = self.net.forward(self.output_layers)
        class_ids = []
        confidences = []
        boxes = []
        center = []
        for out in outs:
            for detection in out:
                scores = detection[5:]
                class_id = np.argmax(scores)
                confidence = scores[class_id]
                if confidence > 0.1:
                    # object detected
                    center_x = int(detection[0] * width)
                    center_y = int(detection[1] * height)
                    w = int(detection[2] * width)
                    h = int(detection[3] * height)
                    # rectangle coordinates
                    x = int(center_x - w / 2)
                    y = int(center_y - h / 2)
                    center.append([center_x, center_y])
                    boxes.append([x, y, w, h])
                    confidences.append(float(confidence))
                    class_ids.append(class_id)

        values = []
        indexes = cv2.dnn.NMSBoxes(
            boxes, confidences, 0.5, 0.4)  # 0.4 changeable
        font = cv2.FONT_HERSHEY_PLAIN
        for i in range(len(boxes)):
            if i in indexes:
                label = str(self.classes[class_ids[i]])
                x, y, w, h = boxes[i]
                temp = {
                    "class": label,
                    "confidence": confidences[i],
                    "x": x,
                    "y": y,
                    "width": w,
                    "height": h,
                    "center": center[i],
                    "color": self.colors[class_ids[i]]
                }
                values.append(temp)
        return values

    def getClassMapping(self):
        class_mapping = {}
        try:
            with open(self.classes_path, 'r') as file:
                lines = file.readlines()
                for index, line in enumerate(lines):
                    class_name = line.strip()
                    class_mapping[class_name] = index
        except FileNotFoundError:
            print(f"Error: File '{self.classes_path}' not found.")
        except Exception as e:
            print(f"Error: An unexpected error occurred - {e}")
        return class_mapping


class ImgBuster(ImgRex):  # 5
    def __init__(self):
        self.classes = None
        self.colors = None
        self.model = None
        self.count = 0

    def load(self, names, weight):
        with open(names, "r") as name:
            self.classes = name.read().split("\n")
        self.colors = np.random.uniform(0, 255, size=(len(self.classes), 3))
        # print(f"self.colors[0] = {self.colors[0]}, type = {type(self.colors)}")
        # self.model = torch.hub.load('ultralytics/yolov5', 'custom', weight, force_reload=True)
        # self.model = torch.hub.load('yolov5', 'custom', weight, source='local')

        count = 0
        success = False
        max_count = 100
        while not success:
            print(f"[INFO] connecting {count} ...")
            try:
                self.model = torch.hub.load(
                    'ultralytics/yolov5', 'custom', weight)
                success = True
            except urllib.error.URLError as e:
                print(f"[ERROR] {e}")
                time.sleep(10.0)
                count += 1
        if not success:
            print(f"[ERROR] Connection not stable error code: {max_count}!!")

    def predict(self, frame):
        width = frame.shape[1]
        height = frame.shape[0]

        values = []
        results = self.model(frame)
        pred = results.pred[0]
        boxes_t = pred[:, :4].cpu().numpy()
        labels_t = pred[:, -1].cpu().numpy()
        confidences_t = pred[:, 4].cpu().numpy()

        boxes = []
        center = []
        class_ids = []
        confidences = []
        try:
            # frame = np.squeeze(results.render())
            for box, label, confidence in zip(boxes_t, labels_t, confidences_t):
                if confidence > 0.1:
                    x1, y1, x2, y2 = box
                    boxes.append([int(x1), int(y1), int(
                        x2) - int(x1), int(y2) - int(y1)])
                    center.append([int((x1 + x2) / 2), int((y1 + y2) / 2)])
                    confidences.append(round(confidence, 2))
                    class_ids.append(int(label))

            for i in range(len(boxes)):
                x, y, w, h = boxes[i]
                temp = {
                    "class": str(self.classes[class_ids[i]]),
                    "confidence": confidences[i],
                    "x": x,
                    "y": y,
                    "width": w,
                    "height": h,
                    "center": center[i],
                    "color": self.colors[class_ids[i]]
                }
                values.append(temp)
        except TypeError:
            pass

        return values


class HogDescriptor:
    def __init__(self):
        self.hog = cv2.HOGDescriptor()
        self.hog.setSVMDetector(cv2.HOGDescriptor_getDefaultPeopleDetector())

    def predict(self, frame):
        values = []
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        boxes, weights = self.hog.detectMultiScale(
            gray, winStride=(8, 8), padding=(32, 32), scale=1.05)
        for (x, y, w, h) in boxes:
            temp = {
                "class": "person",
                "confidence": 0.5,
                "x": x,
                "y": y,
                "width": w,
                "height": h,
                "center": [(x + w) // 2, (y + h) // 2],
                "color": np.array([255.12, 10.22, 20.3])
            }
            values.append(temp)
        return values


class FingerprintController:

    def __init__(self):
        self.uart = None
        self.finger = None
        self.id_name_map = {}

    def initialize(self, uart_port="COM7", baud_rate=57600):
        self.uart = serial.Serial(uart_port, baudrate=baud_rate, timeout=1)
        self.finger = adafruit_fingerprint.Adafruit_Fingerprint(self.uart)

    # def get_fingerprint(self):
    #     print("Waiting for image...")
    #     while self.finger.get_image() != adafruit_fingerprint.OK:
    #         pass
    #     print("Templating...")
    #     if self.finger.image_2_tz(1) != adafruit_fingerprint.OK:
    #         return False
    #     print("Searching...")
    #     if self.finger.finger_search() != adafruit_fingerprint.OK:
    #         return False
    #     return True
    def get_fingerprint(self):
        print("Waiting for image...")
        while self.finger.get_image() != adafruit_fingerprint.OK:
            pass
        print("Templating...")
        if self.finger.image_2_tz(1) != adafruit_fingerprint.OK:
            return False
        print("Searching...")
        if self.finger.finger_search() != adafruit_fingerprint.OK:
            return False
        # Inisialisasi
        detected_id = self.finger.finger_id  # Get the detected fingerprint's ID
        if detected_id in self.id_name_map:
            detected_name = self.id_name_map[detected_id]  # Get the name associated with the ID
        else:
            detected_name = "Nama Tidak Teregister"  # Use "Unknown" as the name if the ID is not registered
        # Save the fingerprint image as a PNG file with the format "ID_Name.png"
        filename = f"{detected_name}_{detected_id}.png"
        while self.finger.get_image():
            pass
        image_saved = self.save_fingerprint_image(filename)
        if image_saved:
            print(f"Fingerprint image saved as {filename}")
        else:
            print("Failed to save fingerprint image")

        return True

    def get_fingerprint_self(self):
        while True:
            print("Waiting for image...")
            while self.finger.get_image() != adafruit_fingerprint.OK:
                pass
            print("Templating...")
            if self.finger.image_2_tz(1) != adafruit_fingerprint.OK:
                return False
            print("Searching...")
            if self.finger.finger_search() != adafruit_fingerprint.OK:
                return False
            # Inisialisasi
            detected_id = self.finger.finger_id  # Get the detected fingerprint's ID
            if detected_id in self.id_name_map:
                detected_name = self.id_name_map[detected_id]  # Get the name associated with the ID
            else:
                detected_name = "Nama Tidak Teregister"  # Use "Unknown" as the name if the ID is not registered

            # Save the fingerprint image as a PNG file with a unique format
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
            filename = f"{detected_name}_{detected_id}_{timestamp}.png"
            while self.finger.get_image():
                pass

            image_saved = self.save_fingerprint_image(filename)
            if image_saved:
                print(f"Fingerprint image saved as {filename}")
            else:
                print("Failed to save fingerprint image")

    def get_fingerprint_self_zebra(self):
        # while True:
        import json
        import time

        last_warn_time = 0
        warn_time = 1500
        now = time.time()
        print("Waiting for image...")

        detected_name = 'zebra'

        detected_id = 0
        # Save the fingerprint image as a PNG file with a unique format
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        filename = f"{detected_name}_{detected_id}_{timestamp}.png"

        data = {
            'detek': False,
            'file': False,
        }
        json_data = json.dumps(data)
        try:
            while self.finger.get_image() != adafruit_fingerprint.OK:
                pass
            image_saved = self.save_fingerprint_image(filename)
            if image_saved:
                data = {
                    'detek': True,
                    'file': filename,
                }

                json_data = json.dumps(data)
                print(json_data)
                # print(f"Fingerprint image saved as {filename}")
            else:
                print("Failed to save fingerprint image")

        except:
            data = {
                'detek': False,
                'file': False,
            }
            json_data = json.dumps(data)
            print(json_data)

    def get_fingerprint_detail(self):
        finger_condition = 0  # Inisialisasi kondisi sidik jari

        print("Getting image...", end="")
        i = self.finger.get_image()
        if i == adafruit_fingerprint.OK:
            print("Image taken")
        else:
            if i == adafruit_fingerprint.NOFINGER:
                print("No finger detected")
            elif i == adafruit_fingerprint.IMAGEFAIL:
                print("Imaging error")
            else:
                print("Other error")
            return False

        print("Templating...", end="")
        i = self.finger.image_2_tz(1)
        if i == adafruit_fingerprint.OK:
            print("Templated")
        else:
            if i == adafruit_fingerprint.IMAGEMESS:
                print("Image too messy")
            elif i == adafruit_fingerprint.FEATUREFAIL:
                print("Could not identify features")
            elif i == adafruit_fingerprint.INVALIDIMAGE:
                print("Image invalid")
            else:
                print("Other error")
            return False

        print("Searching...", end="")
        i = self.finger.finger_fast_search()
        if i == adafruit_fingerprint.OK:
            print("Found fingerprint!")
            finger_condition = 1
        elif i == adafruit_fingerprint.NOTFOUND:
            print("No match found")
            finger_condition = 0
        else:
            print("Other error")

        return finger_condition

    def enroll_finger(self, location):
        for fingerimg in range(1, 3):
            if fingerimg == 1:
                print("Place finger on sensor...", end="")
            else:
                print("Place same finger again...", end="")

            while True:
                i = self.finger.get_image()
                if i == adafruit_fingerprint.OK:
                    print("Image taken")
                    break
                if i == adafruit_fingerprint.NOFINGER:
                    print(".", end="")
                elif i == adafruit_fingerprint.IMAGEFAIL:
                    print("Imaging error")
                    return False
                else:
                    print("Other error")
                    return False

            print("Templating...", end="")
            i = self.finger.image_2_tz(fingerimg)
            if i == adafruit_fingerprint.OK:
                print("Templated")
            else:
                if i == adafruit_fingerprint.IMAGEMESS:
                    print("Image too messy")
                elif i == adafruit_fingerprint.FEATUREFAIL:
                    print("Could not identify features")
                elif i == adafruit_fingerprint.INVALIDIMAGE:
                    print("Image invalid")
                else:
                    print("Other error")
                return False

            if fingerimg == 1:
                print("Remove finger")
                time.sleep(1)
                while i != adafruit_fingerprint.NOFINGER:
                    i = self.finger.get_image()

        print("Creating model...", end="")
        i = self.finger.create_model()
        if i == adafruit_fingerprint.OK:
            print("Created")
        else:
            if i == adafruit_fingerprint.ENROLLMISMATCH:
                print("Prints did not match")
            else:
                print("Other error")
            return False

        print("Storing model #%d..." % location, end="")
        i = self.finger.store_model(location)
        if i == adafruit_fingerprint.OK:
            print("Stored")
        else:
            if i == adafruit_fingerprint.BADLOCATION:
                print("Bad storage location")
            elif i == adafruit_fingerprint.FLASHERR:
                print("Flash storage error")
            else:
                print("Other error")
            return False

        return True

    def save_fingerprint_image(self, filename):
        while self.finger.get_image():
            pass

        img = Image.new("L", (256, 288), "white")
        pixeldata = img.load()
        mask = 0b00001111
        result = self.finger.get_fpdata(sensorbuffer="image")
        x = 0
        y = 0
        for i in range(len(result)):
            pixeldata[x, y] = (int(result[i]) >> 4) * 17
            x += 1
            pixeldata[x, y] = (int(result[i]) & mask) * 17
            if x == 255:
                x = 0
                y += 1
            else:
                x += 1

        datasets_path = f"213/{filename}"  # Path to the datasets directory
        if not img.save(datasets_path):
            return True
        return False

    def get_num(self, max_number):
        i = -1
        while (i > max_number - 1) or (i < 0):
            try:
                i = int(input("Enter ID # from 0-{}: ".format(max_number - 1)))
                if i in self.id_name_map:
                    print(f"Name for ID #{i}: {self.id_name_map[i]}")
                else:
                    print(f"No name registered for ID #{i}")
                    register_name = input("Enter a name for ID #{}: ".format(i))
                    self.id_name_map[i] = register_name
            except ValueError:
                pass
        return i

    def predictFinger(self):
        finger_detected = False
        while True:
            print("Fingerprint Recognition")
            print("----------------")
            print("e) Training Fingerprint")
            print("f) Deteksi Fingerprint")
            print("d) Delete User pada Fingerprint")
            print("s) Save Gambar Fingerprint")
            # print("r) Reset Library")
            print("q) Continue to Program Face Recognition")
            print("x) Exit Fingerprint Recognition Program")
            print("Z) Zebra")
            c = input("> ")

            if c == "e":
                self.enroll_finger(self.get_num(self.finger.library_size))
            # elif c == "f":
            #     if self.get_fingerprint():
            #         detected_name = self.finger.finger_id
            #         print("Detected #", self.finger.finger_id, self.id_name_map[detected_name], "with confidence",
            #               self.finger.confidence)
            #         finger_detected = True
            #     else:
            #         print("Finger not found")
            #         finger_detected = False
            elif c == "f":
                if self.get_fingerprint_self():
                    detected_id = self.finger.finger_id  # Get the detected fingerprint's ID
                    if detected_id in self.id_name_map:
                        detected_name = self.id_name_map[detected_id]  # Get the name associated with the ID
                    else:
                        detected_name = "Name not register"  # Use "Unknown" as the name if the ID is not registered

                    print("Detected #", detected_id, detected_name, "with confidence", self.finger.confidence)
                    finger_detected = True
                else:
                    print("Finger not found")
                    finger_detected = False

            elif c == "z":
                if self.get_fingerprint_self_zebra():
                    detected_id = self.finger.finger_id  # Get the detected fingerprint's ID
                    if detected_id in self.id_name_map:
                        detected_name = self.id_name_map[detected_id]  # Get the name associated with the ID
                    else:
                        detected_name = "Zebra Name not register"  # Use "Unknown" as the name if the ID is not registered

                    print("Zebra Detected #", detected_id, detected_name, "Zebra with confidence",
                          self.finger.confidence)
                    finger_detected = True
                else:
                    print("Zebra Finger not found")
                    finger_detected = False

            elif c == "d":
                location = self.get_num(self.finger.library_size)
                if self.finger.delete_model(location) == adafruit_fingerprint.OK:
                    if location in self.id_name_map:
                        deleted_name = self.id_name_map[location]
                        del self.id_name_map[location]
                        print(f"Deleted ID #{location} and Name ID:", deleted_name)
                    print("Deleted!")
                else:
                    print("Failed to delete")
            elif c == "s":
                if self.save_fingerprint_image("fingerprint.png"):
                    print("Fingerprint image saved")
                else:
                    print("Failed to save fingerprint image")
            elif c == "r":
                if self.finger.empty_library() == adafruit_fingerprint.OK:
                    print("Library empty!")
                else:
                    print("Failed to empty library")
            elif c == "q":
                print("Exiting fingerprint program")
                break
            elif c == "x":
                raise SystemExit
            else:
                print("Invalid option. Please select a valid option.")

        return finger_detected


class FaceRecognitionTraining:
    def __init__(self):
        self.recognizer = cv2.face.LBPHFaceRecognizer_create()
        self.detector = cv2.CascadeClassifier(
            "assets/cascades/data/haarcascades/haarcascade_frontalface_default.xml")
        self.is_train = False

    def getImagesWithLabels(self, path):
        imagePaths = [os.path.join(path, f) for f in os.listdir(path)]
        faceSamples = []
        Ids = []
        for imagePath in imagePaths:
            pilImage = Image.open(imagePath).convert('L')
            imageNp = np.array(pilImage, 'uint8')
            Id = int(os.path.split(imagePath)[-1].split(".")[1])
            faces = self.detector.detectMultiScale(imageNp)
            for (x, y, w, h) in faces:
                faceSamples.append(imageNp[y:y + h, x:x + w])
                Ids.append(Id)
        return faceSamples, Ids

    def process(self):
        print(f"[INFO] Started Training")
        faces, Ids = self.getImagesWithLabels('assets/cascades/datasets')
        self.recognizer.train(faces, np.array(Ids))
        self.recognizer.save('assets/cascades/training/training.xml')
        print(f"[INFO] Training Finished")


class FaceRecognition:
    def __init__(self):
        self.gray = None
        self.frame = None
        self.temp_names = None
        self.frame_detect = None
        self.detect = cv2.CascadeClassifier(
            "assets/cascades/data/haarcascades/haarcascade_frontalface_default.xml")

        self.recognizer = cv2.face.LBPHFaceRecognizer_create()
        self.recognizer.read("assets/cascades/training/training.xml")
        self.name_to_save = ""

        self.read_name = []
        self.names = ""
        with open("assets/cascades/names/name.names") as f:
            self.names = f.readlines()

        self.read_name = [x.strip() for x in self.names]
        self.result_name = [x.split(" ") for x in self.read_name]

        self.coordinate = [0, 0, 0, 0]
        self.capture_save = False
        self.max_samples = 0
        self.stop = False
        self.u_time = 0
        self.name = ""
        self.index = 0
        self.id = 0

    def setMaxNumSamples(self, samples=500):
        self.max_samples = samples

    def showUser(self, name_path="assets/cascades/names/name.names"):
        with open(name_path, 'r') as file:
            lines = file.readlines()
        id_user = []
        name_user = []
        for line in lines:
            data = line.strip().split(' ')
            id_user.append(data[0])
            name_user.append(' '.join(data[1:]))
        df = pd.DataFrame({'ID': id_user, 'Nama': name_user})
        print(df)

    def deleteAllUser(self, path="assets/cascades/datasets", name_path="assets/cascades/names/name.names"):
        extension = ['.jpg', '.jpeg', '.png', '.gif', '.bmp']
        image_list = []
        for ext in extension:
            image_list.extend(glob.glob(os.path.join(path, f"*{ext}")))
        for image in image_list:
            try:
                os.remove(image)
                print(f"[INFO] Image Deleted {image}")
            except Exception as e:
                print(f"[INFO] Failed to Delete {image}: {e}")
        with open(name_path, 'w') as file:
            file.write('')
        print(f"[INFO] Success Delete All User")

    def deleteUser(self, user_number=-1, path="assets/cascades/datasets", name_path="assets/cascades/names/name.names"):
        extension = ['.jpg', '.jpeg', '.png', '.gif', '.bmp']
        pattern = os.path.join(path, f"user.{user_number}.*")
        image_list = []
        for ext in extension:
            image_list.extend(glob.glob(f"{pattern}{ext}"))
        for image in image_list:
            try:
                os.remove(image)
                print(f"[INFO] Image Deleted {image}")
            except Exception as e:
                print(f"[INFO] Failed to Delete {image}: {e}")
        with open(name_path, 'r') as file:
            lines = file.readlines()
        lines = [line for line in lines if not line.startswith(f"{user_number} ")]
        with open(name_path, 'w') as file:
            file.writelines(lines)
        print(f"[INFO] Success Delete User: {user_number}")

    def predict(self, frame):
        try:
            self.gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            self.frame_detect = self.detect.detectMultiScale(self.gray, 1.3, 5)
            if self.capture_save and len(self.frame_detect) > 0:
                self.index += 1
                print(f"[INFO] User {self.id} Image Saved {self.index}")
            values = []
            for (x, y, w, h) in self.frame_detect:
                id, conf = self.recognizer.predict(self.gray[y:y + h, x:x + w])
                for i in range(len(self.result_name)):
                    for j in range(len(self.result_name[i])):
                        if self.result_name[i][0] == str(id):
                            # self.temp_names = f"{self.result_name[i][0]}|{self.result_name[i][1]}"
                            self.temp_names = f"{self.result_name[i][1]}"
                if self.temp_names is not None:
                    temp = {
                        "class": self.temp_names,
                        "confidence": conf / 100,
                        "x": x,
                        "y": y,
                        "width": w,
                        "height": h,
                        "center": [(x + w) // 2, (y + h) // 2],
                        "color": np.array([255.12, 10.22, 20.3])
                    }
                    values.append(temp)
                if self.capture_save:
                    cv2.imwrite("assets/cascades/datasets/user." +
                                str(self.id) + "." + str(self.index) + ".jpg",
                                self.gray[y:y + h, x: x + w])
                    self.name_to_save = f"{str(self.id)} {self.name}\n"
                    if not self.index > (self.max_samples - 1):
                        self.stop = False
                    else:
                        self.stop = True
                        with open("assets/cascades/names/name.names", "a+") as f:
                            f.write(self.name_to_save)
            return values
        except TypeError as e:
            print(e)

    def isStop(self):
        return self.stop

    def isTraining(self):
        return self.capture_save

    def getCoordinate(self):
        return self.coordinate

    def Capture(self, state=False):
        self.capture_save = state
        if state:
            while True:
                try:
                    self.id = int(input("[INFO] Input your ID's   : "))
                    with open("assets/cascades/names/name.names", 'r') as file:
                        lines = file.readlines()
                    for line in lines:
                        if line.startswith(f"{self.id} "):
                            raise Exception()
                except ValueError as e:
                    print(f"\033[91m[ERROR] ID's must be an Number \n{e}.\033[0m")
                    continue
                except Exception as e:
                    print(f"\033[91m[ERROR] ID User '{self.id}' is Used, Pls Use Other ID.\033[0m")
                    continue
                try:
                    self.name = str(input("[INFO] Input your Name's : "))
                    if self.isContainsNum(self.name) or self.name == "":
                        raise ValueError()
                    break
                except ValueError as e:
                    print(f"\033[91m[ERROR] Name's must be an Character.\033[0m")
            print("\033[92m[INFO] Input ID's Success..\033[0m")
            print("[INFO] Face the camera")
            count = 5
            while True:
                print(f"[INFO] Face the camera [ {count} ]")
                if count <= 1:
                    break
                else:
                    count -= 1
                time.sleep(1)

    def getIDs(self):
        return self.id

    def isContainsNum(self, str_in):
        return any(char.isdigit() for char in str_in)


class HOGDescriptor(ImgRex):  # 5
    def __init__(self):
        self.classes = None
        self.colors = None
        self.model = None
        self.count = 0

    def load(self, names="assets/class/coco.txt", weight="assets/models/coco-yolo5.pt"):
        with open(names, "r") as name:
            self.classes = name.read().split("\n")
        self.colors = np.random.uniform(0, 255, size=(len(self.classes), 3))
        # print(f"self.colors[0] = {self.colors[0]}, type = {type(self.colors)}")
        # self.model = torch.hub.load('ultralytics/yolov5', 'custom', weight, force_reload=True)
        # self.model = torch.hub.load('yolov5', 'custom', weight, source='local')

        count = 0
        success = False
        max_count = 100
        while not success:
            print(f"[INFO] connecting {count} ...")
            try:
                self.model = torch.hub.load(
                    'ultralytics/yolov5', 'custom', weight)
                success = True
            except urllib.error.URLError as e:
                print(f"[ERROR] {e}")
                time.sleep(10.0)
                count += 1
        if not success:
            print(f"[ERROR] Connection not stable error code: {max_count}!!")

    def predict(self, frame):
        width = frame.shape[1]
        height = frame.shape[0]

        values = []
        results = self.model(frame)
        pred = results.pred[0]
        boxes_t = pred[:, :4].cpu().numpy()
        labels_t = pred[:, -1].cpu().numpy()
        confidences_t = pred[:, 4].cpu().numpy()

        boxes = []
        center = []
        class_ids = []
        confidences = []
        try:
            # frame = np.squeeze(results.render())
            for box, label, confidence in zip(boxes_t, labels_t, confidences_t):
                if confidence > 0.1:
                    x1, y1, x2, y2 = box
                    boxes.append([int(x1), int(y1), int(
                        x2) - int(x1), int(y2) - int(y1)])
                    center.append([int((x1 + x2) / 2), int((y1 + y2) / 2)])
                    confidences.append(round(confidence, 2))
                    class_ids.append(int(label))

            for i in range(len(boxes)):
                x, y, w, h = boxes[i]
                temp = {
                    "class": str(self.classes[class_ids[i]]),
                    "confidence": confidences[i],
                    "x": x,
                    "y": y,
                    "width": w,
                    "height": h,
                    "center": center[i],
                    "color": self.colors[class_ids[i]]
                }
                values.append(temp)
        except TypeError:
            pass

        return values

# @app.route('/absensi', methods='GET')
# def viewAbsensi():
#     return render_template('absensi.html')
#
# @app.route('/absensi', methods='POST')
# def doAbsensi():
#     nim = request.form['nim']
#
#     # request ke server
#     res = requests.get('https://host.com/cekNIM/' + nim)
#     valid = #parsing jsom(res)
#
#     if valid is True:
#         # isFingerValid, imFinger = True, 'tangan.jpg'
#         isFingerValid, imFinger = getFingerImageData()
#         if isFingerValid is True :
#             files = {'media': imFinger}
#             res = requests.get('https://host.com/absensi/' + nim, gambar=files)
#             cekRespon = # parsing json(res)
#
#             if cekRespon['valid']:
#                 print(absen masuk)
#             else :
#                 print anda sudah absen



import os
import cv2
import time

class FingerprintMatcher:
    def __init__(self, sample_image_path):
        self.sample = cv2.imread(sample_image_path)
        if self.sample is None:
            raise ValueError("Gambar sample tidak valid.")

    def match(self, dataset_folder):
        start_time = time.time()
        best_score = 0
        filename = None
        image = None
        kp1, kp2, mp = None, None, None

        for file in [file for file in os.listdir(dataset_folder)][:1000]:
            fingerprint_image = cv2.imread(os.path.join(dataset_folder, file))
            sift = cv2.SIFT_create()

            keypoints_1, descriptor_1 = sift.detectAndCompute(self.sample, None)
            keypoints_2, descriptor_2 = sift.detectAndCompute(fingerprint_image, None)

            bf = cv2.BFMatcher(cv2.NORM_L2)
            matches = bf.knnMatch(descriptor_1, descriptor_2, k=2)
            match_points = []

            for p, q in matches:
                if p.distance < 0.3 * q.distance:  # maks 0.3
                    match_points.append(p)

            keypoints = min(len(keypoints_1), len(keypoints_2))
            print(keypoints)
            if keypoints > 0 and len(match_points) / keypoints * 100 > best_score:
                best_score = len(match_points) / keypoints * 100
                filename = file
                image = fingerprint_image
                kp1, kp2, mp = keypoints_1, keypoints_2, match_points

        if filename is not None:
            print("BEST MATCH: " + filename)
        else:
            print("Tidak Ditemukan")

        print("HASIL DETEKSI: " + str(best_score))
        print("TIME DELAY: ", time.time() - start_time)

        if self.sample is not None and image is not None:
            result = cv2.drawMatches(self.sample, kp1, image, kp2, mp, None)
            result = cv2.resize(result, (result.shape[1] * 2, result.shape[0] * 2))
            cv2.imshow("Hasil", result)
            cv2.waitKey(0)
            cv2.destroyAllWindows()
        else:
            print("Gambar sample atau image tidak valid.")



import os
import bz2
import gdown
import numpy as np
from deepface.commons import functions

# pylint: disable=too-few-public-methods


class DlibResNet:
    def __init__(self):

        # this is not a must dependency
        import dlib  # 19.20.0

        self.layers = [DlibMetaData()]

        # ---------------------

        home = functions.get_deepface_home()
        weight_file = home + "/.deepface/weights/dlib_face_recognition_resnet_model_v1.dat"

        # ---------------------

        # download pre-trained model if it does not exist
        if os.path.isfile(weight_file) != True:
            print("dlib_face_recognition_resnet_model_v1.dat is going to be downloaded")

            file_name = "dlib_face_recognition_resnet_model_v1.dat.bz2"
            url = f"http://dlib.net/files/{file_name}"
            output = f"{home}/.deepface/weights/{file_name}"
            gdown.download(url, output, quiet=False)

            zipfile = bz2.BZ2File(output)
            data = zipfile.read()
            newfilepath = output[:-4]  # discard .bz2 extension
            with open(newfilepath, "wb") as f:
                f.write(data)

        # ---------------------

        model = dlib.face_recognition_model_v1(weight_file)
        self.__model = model

        # ---------------------

        # return None  # classes must return None

    def predict(self, img_aligned):

        # functions.detectFace returns 4 dimensional images
        if len(img_aligned.shape) == 4:
            img_aligned = img_aligned[0]

        # functions.detectFace returns bgr images
        img_aligned = img_aligned[:, :, ::-1]  # bgr to rgb

        # deepface.detectFace returns an array in scale of [0, 1]
        # but dlib expects in scale of [0, 255]
        if img_aligned.max() <= 1:
            img_aligned = img_aligned * 255

        img_aligned = img_aligned.astype(np.uint8)

        model = self.__model

        img_representation = model.compute_face_descriptor(img_aligned)

        img_representation = np.array(img_representation)
        img_representation = np.expand_dims(img_representation, axis=0)

        return img_representation


class DlibMetaData:
    def __init__(self):
        self.input_shape = [[1, 150, 150, 3]]



import sys
import os

import numpy as np
import cv2
import urllib.request
import time


class Vision:
    def __init__(self, isUsingCam=None, addr=None, index=0):
        # write configuration
        self.frame_count = 0
        self.filenames = None
        self.fourcc = None
        self.out = None

        # get address
        self.cap = None
        self.success = False
        self.index = index
        if isUsingCam:
            while not self.success:
                try:
                    print(f"[INFO] Initialize Camera with Index {self.index}")
                    self.cap = cv2.VideoCapture(self.index)
                    if not self.cap.isOpened():
                        raise Exception(f"Cannot Open Camera by Index {self.index}")
                    ret, frame = self.cap.read()
                    if not ret:
                        raise Exception(f"Failed to Capture Frame by Index {self.index}")
                    self.success = True
                except Exception as err:
                    print(f"[ERROR] Camera Initialization Failed: {err}")
                    time.sleep(1.5)
                    self.index += 1
            print(f"[INFO] Camera Initialization Success")
        else:
            self.cap = cv2.VideoCapture(addr)

        # fps
        self._prev_time = 0
        self._new_time = 0

    def writeConfig(self, name="output.mp4", types="mp4v"):  # XVID -> avi
        self.filenames = name
        self.fourcc = cv2.VideoWriter_fourcc(*types)  # format video
        # filename, format, FPS, frame size
        frame_width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        frame_height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        fps = int(self.cap.get(cv2.CAP_PROP_FPS))
        self.out = cv2.VideoWriter(self.filenames, self.fourcc, fps, (frame_width, frame_height))

    def write(self, frame):
        self.out.write(frame)

    def writeImg(self, frame, path="cats-output.png"):
        filename = path
        cv2.imwrite(filename, frame)
        with open(filename, 'ab') as f:
            f.flush()
            os.fsync(f.fileno())

    def resize(self, image, width=None, height=None,
               interpolation=cv2.INTER_AREA):
        dim = None
        w = image.shape[1]
        h = image.shape[0]
        if width is None and height is None:
            return image
        if width is None:
            r = height / float(h)
            dim = (int(w * r), height)
        else:
            r = width / float(w)
            dim = (width, int(h * r))
        resized = cv2.resize(image, dim, interpolation=interpolation)
        return resized

    def __get_fps(self):
        fps = 0.0
        try:
            self._new_time = time.time()
            fps = 1 / (self._new_time - self._prev_time)
            self._prev_time = self._new_time
            fps = 30 if fps > 30 else 0 if fps < 0 else fps
        except ZeroDivisionError as e:
            pass
        return int(fps)

    def blur(self, frame=None, sigma=11):
        return cv2.GaussianBlur(frame, (sigma, sigma), 0)

    def autoContrast(self, frame):
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        min_val, max_val, _, _ = cv2.minMaxLoc(gray)
        adjusted_image = cv2.convertScaleAbs(frame, alpha=255.0 / (max_val - min_val),
                                             beta=-min_val * 255.0 / (max_val - min_val))
        return adjusted_image

    def adaptiveContrast(self, frame, clip_limit=2.0, tile_grid_size=(8, 8)):
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)
        enhanced_gray = clahe.apply(gray)
        enhanced_image = cv2.cvtColor(enhanced_gray, cv2.COLOR_GRAY2BGR)
        return enhanced_image

    def equalizeHistogram(self, frame):
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        equalized_gray = cv2.equalizeHist(gray)
        equalized_image = cv2.cvtColor(equalized_gray, cv2.COLOR_GRAY2BGR)
        return equalized_image

    def enhanceColors(self, frame, saturation_factor):
        hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
        hsv[..., 1] = hsv[..., 1] * saturation_factor
        hsv[..., 1] = np.clip(hsv[..., 1], 0, 255)
        enhanced_image = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
        return enhanced_image

    def sharpenImage(self, frame, sharp):
        blurred = cv2.GaussianBlur(frame, (0, 0), 3)
        sharpened = cv2.addWeighted(frame, 1 + sharp, blurred, -sharp, 0)
        return sharpened

    def sharpen(self, frame):
        blurred = cv2.GaussianBlur(frame, (0, 0), 3)
        sharpened = cv2.addWeighted(frame, 1.5, blurred, -0.5, 0)
        return sharpened

    def setBrightness(self, frame, value):
        h, s, v = cv2.split(
            cv2.cvtColor(frame, cv2.COLOR_BGR2HSV))
        v = np.clip(v.astype(int) + value, 0, 255).astype(np.uint8)
        return cv2.cvtColor(
            cv2.merge((h, s, v)), cv2.COLOR_HSV2BGR)

    def setContrast(self, frame, value):
        alpha = float(131 * (value + 127)) / (127 * (131 - value))
        gamma = 127 * (1 - alpha)
        return cv2.addWeighted(
            frame, alpha, frame, 0, gamma)

    def setBrightnessNcontrast(self, frame, bright=0.0, contr=0.0, beta=0.0):
        return cv2.addWeighted(frame, 1 + float(contr)
                               / 100.0, frame, beta, float(bright))

    def read(self, frame_size=0, show_fps=False):
        try:
            success, frame = self.cap.read()
            if not success:
                raise RuntimeError
            if show_fps:
                try:  # put fps
                    cv2.putText(frame, str(self.__get_fps()) + " fps", (20, 40), 0, 1,
                                [225, 255, 255], thickness=2, lineType=cv2.LINE_AA)
                except RuntimeError as e:
                    print(e)
            if frame_size != 0:
                return self.resize(frame, frame_size)
            return frame
        except RuntimeError as e:
            print("[INFO] Failed to capture the Frame")

    def readFromUrl(self, url="http://192.168.200.24/cam-hi.jpg", frame_size=480, show_fps=False):
        try:
            img_resp = urllib.request.urlopen(url)
            imgnp = np.array(bytearray(img_resp.read()), dtype=np.uint8)
            frame = cv2.imdecode(imgnp, -1)
            if show_fps:
                try:  # put fps
                    cv2.putText(frame, str(self.__get_fps()) + " fps", (20, 40), 0, 1,
                                [225, 255, 255], thickness=2, lineType=cv2.LINE_AA)
                except RuntimeError as e:
                    print(e)
            frame = self.resize(frame, frame_size)
            return frame
        except RuntimeError as e:
            print("[INFO] Failed to capture the Frame")

    def show(self, frame, winName="frame"):
        cv2.imshow(winName, frame)

    def wait(self, delay):
        return cv2.waitKey(delay)

    def release(self):
        self.cap.release()

    def destroy(self):
        cv2.destroyAllWindows()



import setuptools

# with open("readme.md", "r", encoding="utf-8") as fh:
#     long_description = fh.read()
#
# with open("requirements.txt", "r", encoding="utf-8") as f:
#     requirements = f.read().split("\n")

setuptools.setup(
    name="kadita",
    version="0.0.3",
    author="Firza Ichlasul Amal Ariansyah",
    author_email="napija25@gmail.com",
    description="Computer Vision with DeepFace",
    long_description_content_type="text/markdown",
    url="https://github.com/Kastara-Digital-Technology/KaditaCV",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": ["deepface = deepface.DeepFace:cli"],
    },
    python_requires=">=3.5.5",

)



import os
import gdown
import tensorflow as tf
from deepface.commons import functions

# pylint: disable=unsubscriptable-object

# --------------------------------
# dependency configuration

tf_version = int(tf.__version__.split(".", maxsplit=1)[0])

if tf_version == 1:
    from keras.engine import training
    import keras
    from keras.layers import (
        ZeroPadding2D,
        Input,
        Conv2D,
        BatchNormalization,
        PReLU,
        Add,
        Dropout,
        Flatten,
        Dense,
    )
else:
    from tensorflow.python.keras.engine import training
    from tensorflow import keras
    from tensorflow.keras.layers import (
        ZeroPadding2D,
        Input,
        Conv2D,
        BatchNormalization,
        PReLU,
        Add,
        Dropout,
        Flatten,
        Dense,
    )
# --------------------------------


# url = "https://drive.google.com/uc?id=1LVB3CdVejpmGHM28BpqqkbZP5hDEcdZY"


def loadModel(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/arcface_weights.h5",
):
    base_model = ResNet34()
    inputs = base_model.inputs[0]
    arcface_model = base_model.outputs[0]
    arcface_model = BatchNormalization(momentum=0.9, epsilon=2e-5)(arcface_model)
    arcface_model = Dropout(0.4)(arcface_model)
    arcface_model = Flatten()(arcface_model)
    arcface_model = Dense(512, activation=None, use_bias=True, kernel_initializer="glorot_normal")(
        arcface_model
    )
    embedding = BatchNormalization(momentum=0.9, epsilon=2e-5, name="embedding", scale=True)(
        arcface_model
    )
    model = keras.models.Model(inputs, embedding, name=base_model.name)

    # ---------------------------------------
    # check the availability of pre-trained weights

    home = functions.get_deepface_home()

    file_name = "arcface_weights.h5"
    output = home + "/.deepface/weights/" + file_name

    if os.path.isfile(output) != True:

        print(file_name, " will be downloaded to ", output)
        gdown.download(url, output, quiet=False)

    # ---------------------------------------

    model.load_weights(output)

    return model


def ResNet34():

    img_input = Input(shape=(112, 112, 3))

    x = ZeroPadding2D(padding=1, name="conv1_pad")(img_input)
    x = Conv2D(
        64, 3, strides=1, use_bias=False, kernel_initializer="glorot_normal", name="conv1_conv"
    )(x)
    x = BatchNormalization(axis=3, epsilon=2e-5, momentum=0.9, name="conv1_bn")(x)
    x = PReLU(shared_axes=[1, 2], name="conv1_prelu")(x)
    x = stack_fn(x)

    model = training.Model(img_input, x, name="ResNet34")

    return model


def block1(x, filters, kernel_size=3, stride=1, conv_shortcut=True, name=None):
    bn_axis = 3

    if conv_shortcut:
        shortcut = Conv2D(
            filters,
            1,
            strides=stride,
            use_bias=False,
            kernel_initializer="glorot_normal",
            name=name + "_0_conv",
        )(x)
        shortcut = BatchNormalization(
            axis=bn_axis, epsilon=2e-5, momentum=0.9, name=name + "_0_bn"
        )(shortcut)
    else:
        shortcut = x

    x = BatchNormalization(axis=bn_axis, epsilon=2e-5, momentum=0.9, name=name + "_1_bn")(x)
    x = ZeroPadding2D(padding=1, name=name + "_1_pad")(x)
    x = Conv2D(
        filters,
        3,
        strides=1,
        kernel_initializer="glorot_normal",
        use_bias=False,
        name=name + "_1_conv",
    )(x)
    x = BatchNormalization(axis=bn_axis, epsilon=2e-5, momentum=0.9, name=name + "_2_bn")(x)
    x = PReLU(shared_axes=[1, 2], name=name + "_1_prelu")(x)

    x = ZeroPadding2D(padding=1, name=name + "_2_pad")(x)
    x = Conv2D(
        filters,
        kernel_size,
        strides=stride,
        kernel_initializer="glorot_normal",
        use_bias=False,
        name=name + "_2_conv",
    )(x)
    x = BatchNormalization(axis=bn_axis, epsilon=2e-5, momentum=0.9, name=name + "_3_bn")(x)

    x = Add(name=name + "_add")([shortcut, x])
    return x


def stack1(x, filters, blocks, stride1=2, name=None):
    x = block1(x, filters, stride=stride1, name=name + "_block1")
    for i in range(2, blocks + 1):
        x = block1(x, filters, conv_shortcut=False, name=name + "_block" + str(i))
    return x


def stack_fn(x):
    x = stack1(x, 64, 3, name="conv2")
    x = stack1(x, 128, 4, name="conv3")
    x = stack1(x, 256, 6, name="conv4")
    return stack1(x, 512, 3, name="conv5")



import os
import time
import numpy as np
import pandas as pd
import cv2
from deepface import DeepFace
from deepface.commons import functions

# dependency configuration
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


# pylint: disable=too-many-nested-blocks


def analysis(
        db_path,
        model_name="VGG-Face",
        detector_backend="opencv",
        distance_metric="cosine",
        enable_face_analysis=False,
        source=0,
        time_threshold=1,  # jadikan 0 agar opencv biasa
        frame_threshold=1,
):
    # global variables
    text_color = (255, 255, 255)
    pivot_img_size = 112  # face recognition result image
    time_threshold = 0.1
    enable_emotion = False
    enable_age_gender = False
    # ------------------------
    # find custom values for this input set
    target_size = functions.find_target_size(model_name=model_name)
    # ------------------------
    # build models once to store them in the memory
    # otherwise, they will be built after cam started and this will cause delays
    DeepFace.build_model(model_name=model_name)
    print(f"facial recognition model {model_name} is just built")

    # if enable_face_analysis:
    #     # DeepFace.build_model(model_name="Age")
    #     # print("Age model is just built")
    #     # DeepFace.build_model(model_name="Gender")
    #     # print("Gender model is just built")
    #     # DeepFace.build_model(model_name="Emotion")
    #     # print("Emotion model is just built")
    # -----------------------
    # call a dummy find function for db_path once to create embeddings in the initialization
    DeepFace.find(
        img_path=np.zeros([224, 224, 3]),
        db_path=db_path,
        model_name=model_name,
        detector_backend=detector_backend,
        distance_metric=distance_metric,
        enforce_detection=False,
    )
    # -----------------------
    # visualization
    freeze = False
    face_detected = False
    face_included_frames = 0  # freeze screen if face detected sequantially 5 frames
    freezed_frame = 0
    tic = time.time()

    cap = cv2.VideoCapture(source)  # webcam
    while True:
        _, img = cap.read()

        if img is None:
            break

        # cv2.namedWindow('img', cv2.WINDOW_FREERATIO)
        # cv2.setWindowProperty('img', cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)

        raw_img = img.copy()
        resolution_x = img.shape[1]
        resolution_y = img.shape[0]

        if freeze == False:
            try:
                # just extract the regions to highlight in webcam
                face_objs = DeepFace.extract_faces(
                    img_path=img,
                    target_size=target_size,
                    detector_backend=detector_backend,
                    enforce_detection=False,
                )
                faces = []
                for face_obj in face_objs:
                    facial_area = face_obj["facial_area"]
                    faces.append(
                        (
                            facial_area["x"],
                            facial_area["y"],
                            facial_area["w"],
                            facial_area["h"],
                        )
                    )
            except:  # to avoid exception if no face detected
                faces = []

            if len(faces) == 0:
                face_included_frames = 0
        else:
            faces = []

        detected_faces = []
        face_index = 0
        for x, y, w, h in faces:
            if w > 130:  # discard small detected faces

                face_detected = True
                if face_index == 0:
                    face_included_frames = (
                            face_included_frames + 1
                    )  # increase frame for a single face

                cv2.rectangle(
                    img, (x, y), (x + w, y + h), (0, 255, 0), 2
                )  # draw rectangle to main image

                # cv2.putText(
                #     img,
                #     str(frame_threshold - face_included_frames),
                #     (int(x + w / 4), int(y + h / 1.5)),
                #     cv2.FONT_HERSHEY_SIMPLEX,
                #     4,
                #     (255, 255, 255),
                #     2,
                # )

                detected_face = img[int(y): int(y + h), int(x): int(x + w)]  # crop detected face

                # -------------------------------------

                detected_faces.append((x, y, w, h))
                face_index = face_index + 1

                # -------------------------------------

        if face_detected == True and face_included_frames == frame_threshold and freeze == False:
            print("[INFO] Berhasil Terdeteksi")
            freeze = True
            # base_img = img.copy()
            base_img = raw_img.copy()
            detected_faces_final = detected_faces.copy()
            tic = time.time()

        if freeze == True:

            toc = time.time()
            if (toc - tic) < time_threshold:

                if freezed_frame == 0:
                    freeze_img = base_img.copy()
                    # here, np.uint8 handles showing white area issue
                    # freeze_img = np.zeros(resolution, np.uint8)

                    for detected_face in detected_faces_final:
                        x = detected_face[0]
                        y = detected_face[1]
                        w = detected_face[2]
                        h = detected_face[3]

                        cv2.rectangle(
                            freeze_img, (x, y), (x + w, y + h), (67, 67, 67), 1
                        )  # draw rectangle to main image

                        # -------------------------------
                        # extract detected face
                        custom_face = base_img[y: y + h, x: x + w]
                        # -------------------------------
                        # facial attribute analysis

                        if enable_face_analysis == False:
                            print('ANALYS')
                            demographies = DeepFace.analyze(
                                img_path=custom_face,
                                detector_backend=detector_backend,
                                enforce_detection=False,
                                silent=True,
                            )

                            if len(demographies) > 0:
                                # directly access 1st face cos img is extracted already
                                demography = demographies[0]

                                if enable_emotion:
                                    emotion = demography["emotion"]
                                    emotion_df = pd.DataFrame(
                                        emotion.items(), columns=["emotion", "score"]
                                    )
                                    emotion_df = emotion_df.sort_values(
                                        by=["score"], ascending=False
                                    ).reset_index(drop=True)

                                    # background of mood box

                                    # transparency
                                    overlay = freeze_img.copy()
                                    opacity = 0.4

                                    if x + w + pivot_img_size < resolution_x:
                                        # right
                                        cv2.rectangle(
                                            freeze_img
                                            # , (x+w,y+20)
                                            ,
                                            (x + w, y),
                                            (x + w + pivot_img_size, y + h),
                                            (64, 64, 64),
                                            cv2.FILLED,
                                        )

                                        cv2.addWeighted(
                                            overlay, opacity, freeze_img, 1 - opacity, 0, freeze_img
                                        )

                                    elif x - pivot_img_size > 0:
                                        # left
                                        cv2.rectangle(
                                            freeze_img
                                            # , (x-pivot_img_size,y+20)
                                            ,
                                            (x - pivot_img_size, y),
                                            (x, y + h),
                                            (64, 64, 64),
                                            cv2.FILLED,
                                        )

                                        cv2.addWeighted(
                                            overlay, opacity, freeze_img, 1 - opacity, 0, freeze_img
                                        )

                                    for index, instance in emotion_df.iterrows():
                                        current_emotion = instance["emotion"]
                                        emotion_label = f"{current_emotion} "
                                        emotion_score = instance["score"] / 100

                                        bar_x = 35  # this is the size if an emotion is 100%
                                        bar_x = int(bar_x * emotion_score)

                                        if x + w + pivot_img_size < resolution_x:

                                            text_location_y = y + 20 + (index + 1) * 20
                                            text_location_x = x + w

                                            if text_location_y < y + h:
                                                cv2.putText(
                                                    freeze_img,
                                                    emotion_label,
                                                    (text_location_x, text_location_y),
                                                    cv2.FONT_HERSHEY_SIMPLEX,
                                                    0.5,
                                                    (255, 255, 255),
                                                    1,
                                                )

                                                cv2.rectangle(
                                                    freeze_img,
                                                    (x + w + 70, y + 13 + (index + 1) * 20),
                                                    (
                                                        x + w + 70 + bar_x,
                                                        y + 13 + (index + 1) * 20 + 5,
                                                    ),
                                                    (255, 255, 255),
                                                    cv2.FILLED,
                                                )

                                        elif x - pivot_img_size > 0:

                                            text_location_y = y + 20 + (index + 1) * 20
                                            text_location_x = x - pivot_img_size

                                            if text_location_y <= y + h:
                                                cv2.putText(
                                                    freeze_img,
                                                    emotion_label,
                                                    (text_location_x, text_location_y),
                                                    cv2.FONT_HERSHEY_SIMPLEX,
                                                    0.5,
                                                    (255, 255, 255),
                                                    1,
                                                )

                                                cv2.rectangle(
                                                    freeze_img,
                                                    (
                                                        x - pivot_img_size + 70,
                                                        y + 13 + (index + 1) * 20,
                                                    ),
                                                    (
                                                        x - pivot_img_size + 70 + bar_x,
                                                        y + 13 + (index + 1) * 20 + 5,
                                                    ),
                                                    (255, 255, 255),
                                                    cv2.FILLED,
                                                )

                                if enable_age_gender:
                                    apparent_age = demography["age"]
                                    dominant_gender = demography["dominant_gender"]
                                    gender = "M" if dominant_gender == "Man" else "W"
                                    # print(f"{apparent_age} years old {dominant_emotion}")
                                    analysis_report = str(int(apparent_age)) + " " + gender

                                    # -------------------------------

                                    info_box_color = (46, 200, 255)

                                    # top
                                    if y - pivot_img_size + int(pivot_img_size / 5) > 0:

                                        triangle_coordinates = np.array(
                                            [
                                                (x + int(w / 2), y),
                                                (
                                                    x + int(w / 2) - int(w / 10),
                                                    y - int(pivot_img_size / 3),
                                                ),
                                                (
                                                    x + int(w / 2) + int(w / 10),
                                                    y - int(pivot_img_size / 3),
                                                ),
                                            ]
                                        )

                                        cv2.drawContours(
                                            freeze_img,
                                            [triangle_coordinates],
                                            0,
                                            info_box_color,
                                            -1,
                                        )

                                        cv2.rectangle(
                                            freeze_img,
                                            (
                                                x + int(w / 5),
                                                y - pivot_img_size + int(pivot_img_size / 5),
                                            ),
                                            (x + w - int(w / 5), y - int(pivot_img_size / 3)),
                                            info_box_color,
                                            cv2.FILLED,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            analysis_report,
                                            (x + int(w / 3.5), y - int(pivot_img_size / 2.1)),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            1,
                                            (0, 111, 255),
                                            2,
                                        )

                                    # bottom
                                    elif (
                                            y + h + pivot_img_size - int(pivot_img_size / 5)
                                            < resolution_y
                                    ):

                                        triangle_coordinates = np.array(
                                            [
                                                (x + int(w / 2), y + h),
                                                (
                                                    x + int(w / 2) - int(w / 10),
                                                    y + h + int(pivot_img_size / 3),
                                                ),
                                                (
                                                    x + int(w / 2) + int(w / 10),
                                                    y + h + int(pivot_img_size / 3),
                                                ),
                                            ]
                                        )

                                        cv2.drawContours(
                                            freeze_img,
                                            [triangle_coordinates],
                                            0,
                                            info_box_color,
                                            -1,
                                        )

                                        cv2.rectangle(
                                            freeze_img,
                                            (x + int(w / 5), y + h + int(pivot_img_size / 3)),
                                            (
                                                x + w - int(w / 5),
                                                y + h + pivot_img_size - int(pivot_img_size / 5),
                                            ),
                                            info_box_color,
                                            cv2.FILLED,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            analysis_report,
                                            (x + int(w / 3.5), y + h + int(pivot_img_size / 1.5)),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            1,
                                            (0, 111, 255),
                                            2,
                                        )

                        # --------------------------------
                        # face recognition
                        # call find function for custom_face

                        dfs = DeepFace.find(
                            img_path=custom_face,
                            db_path=db_path,
                            model_name=model_name,
                            detector_backend=detector_backend,
                            distance_metric=distance_metric,
                            enforce_detection=False,
                            silent=True,
                        )

                        if len(dfs) > 0:
                            # directly access 1st item because custom face is extracted already
                            df = dfs[0]

                            if df.shape[0] > 0:
                                candidate = df.iloc[0]
                                label = candidate["identity"]

                                # to use this source image as is
                                display_img = cv2.imread(label)
                                # to use extracted face
                                source_objs = DeepFace.extract_faces(
                                    img_path=label,
                                    target_size=(pivot_img_size, pivot_img_size),
                                    detector_backend=detector_backend,
                                    enforce_detection=False,
                                    align=False,
                                )

                                if len(source_objs) > 0:
                                    # extract 1st item directly
                                    source_obj = source_objs[0]
                                    display_img = source_obj["face"]
                                    display_img *= 255
                                    display_img = display_img[:, :, ::-1]
                                # --------------------
                                label = label.split("/")[-1]
                                # yield  label
                                nama = label.split('.')[0]
                                if '_' in nama:
                                    nama = nama.split('_')[0]
                                print("TERDETEKSI : ", nama)
                                try:
                                    if (
                                            y - pivot_img_size > 0
                                            and x + w + pivot_img_size < resolution_x
                                    ):
                                        # top right
                                        freeze_img[
                                        y - pivot_img_size: y,
                                        x + w: x + w + pivot_img_size,
                                        ] = display_img

                                        overlay = freeze_img.copy()
                                        opacity = 0.4
                                        cv2.rectangle(
                                            freeze_img,
                                            (x + w, y),
                                            (x + w + pivot_img_size, y + 20),
                                            (46, 200, 255),
                                            cv2.FILLED,
                                        )
                                        cv2.addWeighted(
                                            overlay,
                                            opacity,
                                            freeze_img,
                                            1 - opacity,
                                            0,
                                            freeze_img,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            label,
                                            (x + w, y + 10),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            0.5,
                                            text_color,
                                            1,
                                        )

                                        # connect face and text
                                        cv2.line(
                                            freeze_img,
                                            (x + int(w / 2), y),
                                            (x + 3 * int(w / 4), y - int(pivot_img_size / 2)),
                                            (67, 67, 67),
                                            1,
                                        )
                                        cv2.line(
                                            freeze_img,
                                            (x + 3 * int(w / 4), y - int(pivot_img_size / 2)),
                                            (x + w, y - int(pivot_img_size / 2)),
                                            (67, 67, 67),
                                            1,
                                        )

                                    elif (
                                            y + h + pivot_img_size < resolution_y
                                            and x - pivot_img_size > 0
                                    ):
                                        # bottom left
                                        freeze_img[
                                        y + h: y + h + pivot_img_size,
                                        x - pivot_img_size: x,
                                        ] = display_img

                                        overlay = freeze_img.copy()
                                        opacity = 0.4
                                        cv2.rectangle(
                                            freeze_img,
                                            (x - pivot_img_size, y + h - 20),
                                            (x, y + h),
                                            (46, 200, 255),
                                            cv2.FILLED,
                                        )
                                        cv2.addWeighted(
                                            overlay,
                                            opacity,
                                            freeze_img,
                                            1 - opacity,
                                            0,
                                            freeze_img,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            label,
                                            (x - pivot_img_size, y + h - 10),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            0.5,
                                            text_color,
                                            1,
                                        )

                                        # connect face and text
                                        cv2.line(
                                            freeze_img,
                                            (x + int(w / 2), y + h),
                                            (
                                                x + int(w / 2) - int(w / 4),
                                                y + h + int(pivot_img_size / 2),
                                            ),
                                            (67, 67, 67),
                                            1,
                                        )
                                        cv2.line(
                                            freeze_img,
                                            (
                                                x + int(w / 2) - int(w / 4),
                                                y + h + int(pivot_img_size / 2),
                                            ),
                                            (x, y + h + int(pivot_img_size / 2)),
                                            (67, 67, 67),
                                            1,
                                        )

                                    elif y - pivot_img_size > 0 and x - pivot_img_size > 0:
                                        # top left
                                        freeze_img[
                                        y - pivot_img_size: y, x - pivot_img_size: x
                                        ] = display_img

                                        overlay = freeze_img.copy()
                                        opacity = 0.4
                                        cv2.rectangle(
                                            freeze_img,
                                            (x - pivot_img_size, y),
                                            (x, y + 20),
                                            (46, 200, 255),
                                            cv2.FILLED,
                                        )
                                        cv2.addWeighted(
                                            overlay,
                                            opacity,
                                            freeze_img,
                                            1 - opacity,
                                            0,
                                            freeze_img,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            label,
                                            (x - pivot_img_size, y + 10),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            0.5,
                                            text_color,
                                            1,
                                        )

                                        # connect face and text
                                        cv2.line(
                                            freeze_img,
                                            (x + int(w / 2), y),
                                            (
                                                x + int(w / 2) - int(w / 4),
                                                y - int(pivot_img_size / 2),
                                            ),
                                            (67, 67, 67),
                                            1,
                                        )
                                        cv2.line(
                                            freeze_img,
                                            (
                                                x + int(w / 2) - int(w / 4),
                                                y - int(pivot_img_size / 2),
                                            ),
                                            (x, y - int(pivot_img_size / 2)),
                                            (67, 67, 67),
                                            1,
                                        )

                                    elif (
                                            x + w + pivot_img_size < resolution_x
                                            and y + h + pivot_img_size < resolution_y
                                    ):
                                        # bottom righ
                                        freeze_img[
                                        y + h: y + h + pivot_img_size,
                                        x + w: x + w + pivot_img_size,
                                        ] = display_img

                                        overlay = freeze_img.copy()
                                        opacity = 0.4
                                        cv2.rectangle(
                                            freeze_img,
                                            (x + w, y + h - 20),
                                            (x + w + pivot_img_size, y + h),
                                            (46, 200, 255),
                                            cv2.FILLED,
                                        )
                                        cv2.addWeighted(
                                            overlay,
                                            opacity,
                                            freeze_img,
                                            1 - opacity,
                                            0,
                                            freeze_img,
                                        )

                                        cv2.putText(
                                            freeze_img,
                                            label,
                                            (x + w, y + h - 10),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            0.5,
                                            text_color,
                                            1,
                                        )

                                        # connect face and text
                                        cv2.line(
                                            freeze_img,
                                            (x + int(w / 2), y + h),
                                            (
                                                x + int(w / 2) + int(w / 4),
                                                y + h + int(pivot_img_size / 2),
                                            ),
                                            (67, 67, 67),
                                            1,
                                        )
                                        cv2.line(
                                            freeze_img,
                                            (
                                                x + int(w / 2) + int(w / 4),
                                                y + h + int(pivot_img_size / 2),
                                            ),
                                            (x + w, y + h + int(pivot_img_size / 2)),
                                            (67, 67, 67),
                                            1,
                                        )
                                except Exception as err:  # pylint: disable=broad-except
                                    print(str(err))

                        # tic = time.time()  # in this way, freezed image can show 5 seconds

                        # -------------------------------

                time_left = int(time_threshold - (toc - tic) + 1)

                # cv2.rectangle(freeze_img, (10, 10), (90, 50), (67, 67, 67), -10)
                # cv2.putText(
                #     freeze_img,
                #     str(time_left),
                #     (40, 40),
                #     cv2.FONT_HERSHEY_SIMPLEX,
                #     1,
                #     (255, 255, 255),
                #     1,
                # )

                cv2.imshow("img", freeze_img)

                freezed_frame = freezed_frame + 1
            else:
                face_detected = False
                face_included_frames = 0
                freeze = False
                freezed_frame = 0

        else:
            cv2.imshow("img", img)

        if cv2.waitKey(1) & 0xFF == ord("q"):  # press q to quit
            break

    # kill open cv things
    cap.release()
    cv2.destroyAllWindows()



import cv2
from deepface.detectors import FaceDetector


def build_model():
    from mtcnn import MTCNN

    face_detector = MTCNN()
    return face_detector


def detect_face(face_detector, img, align=True):

    resp = []

    detected_face = None
    img_region = [0, 0, img.shape[1], img.shape[0]]

    img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)  # mtcnn expects RGB but OpenCV read BGR
    detections = face_detector.detect_faces(img_rgb)

    if len(detections) > 0:

        for detection in detections:
            x, y, w, h = detection["box"]
            detected_face = img[int(y) : int(y + h), int(x) : int(x + w)]
            img_region = [x, y, w, h]
            confidence = detection["confidence"]

            if align:
                keypoints = detection["keypoints"]
                left_eye = keypoints["left_eye"]
                right_eye = keypoints["right_eye"]
                detected_face = FaceDetector.alignment_procedure(detected_face, left_eye, right_eye)

            resp.append((detected_face, img_region, confidence))

    return resp
